//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	epicbscalgen.cc	--- Compute Boresight misalignment angles for EPIC
//
//	DESCRIPTION:
//		This is an interactive task to compute the boresight misalignment
//		angles for the instruments EMOS1/EMOS2 and EPN from observation
//		data. This used the Nelder/Mead simplex method for minimizing
//		the sum of squares of the differences between a list of
//		catalogued and observed sources positions.
//
//	CHANGE HISTORY:
//		UL, 1999-11-30: 1.0.0: - first version
//		UL, 1999-12-16: 1.1.0: - changed optimization strategy: optimize first
//								 phi/theta up to a tolerance 1.e-3; then
//								 optimize all three angles simulateneously;
//								 this avoid problems with apparent local
//								 minima in psi-parameter direction
//		UL, 1999-01-11: 1.2.0: - name of two parameters have changed:
//								   sourceRadius -> sourceradius
//								   maxIter		-> maxiter
//		UL, 2000-05-03: 1.3.0: - take declination in sum of square calculation
//								 into account
//
//	AUTHOR:
//		U. Lammers, Tue Nov 30 14:34:52 MET 1999
//
//	REFERENCES:
//		[1]: XMM Star Tracker/Instrument Alignment Calibration - Interface
//			 Control Document, Issue 1.2, R. Mugellesi, ESOC/ESA
//			 Online: ftp://astro.estec.esa.nl/pub/XMM/documents/
//					 siam_icd.pdf
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/ccftools/bscalgen/src/epicbscalgen.cc,v 1.9 2001/07/06 08:49:42 ulammers Exp $
//

#include "param.h"
#include "SetTableColumn.h"
#include "CalServer.h"
#include "Boresight.h"
#include "SourceList.h"
#include "Celestial.h"
#include "Numeric.h"
#include "SasTasks.h"

#include "epicbscalgen.h"

//
//	routine to minimize
//
class SourceCorrelator {
public:
	SourceCorrelator(SourceList *slRef, bool runShowDetect = true)
		{ PROC _slRef = slRef; _maxIter = 1000;
			    _runShowDetect = runShowDetect; ENDPROC }

	static void maxIter(unsigned max) { PROC _maxIter = max; ENDPROC }
	static void phi(double phi) { PROC _phi = phi; ENDPROC; }
	static double phi() { PROC RETURN _phi; }
	static void theta(double theta) { PROC _theta = theta; ENDPROC; }
	static double theta() { PROC RETURN _theta; }
	static void psi(double psi) { PROC _psi = psi; ENDPROC; }
	static double psi() { PROC RETURN _psi; }


	static double optimizePhiTheta(double angles[])
	{
		PROC

		_phi = angles[1];
		_theta = angles[2];
		_psi = 0.;
		EulerAngles	eulerAngles(_phi, _theta, _psi, EulerAngles::Euler321);

		RETURN angularSeparation(eulerAngles);
	} // optimizeThetaPhi()


	static double optimizePsi(double angle)
	{
		PROC

		_psi = angle;
		
		EulerAngles	eulerAngles(_phi, _theta, _psi, EulerAngles::Euler321);

		RETURN angularSeparation(eulerAngles);
	} // optimizePsi()


	static double optimizePhiThetaPsi(double angles[])
	{
		PROC

		_phi = angles[1];
		_theta = angles[2];
		_psi = angles[3];
		EulerAngles	eulerAngles(_phi, _theta, _psi, EulerAngles::Euler321);

		RETURN angularSeparation(eulerAngles);
	} // optimizeThetaPhi()


	static double angularSeparation(const EulerAngles &eulerAngles)
	{
		PROC

		//	get new source list with for new misalignment angles
		errstr<<"New set of Euler angles in iteration: "<<_iter++<<" "
			  <<eulerAngles<<endl
			  <<Sas::message(Msg::App, Msg::Silent);

		SourceList sl(_slRef->eventListName(),
					  _slRef->instrument(),
					  _slRef->xAxis(), _slRef->yAxis(),
					  eulerAngles,
					  _slRef->binX(), _slRef->binY(), _slRef->sourceRadius(),
					  _slRef->nruns(), _slRef->likemin(), _runShowDetect);

		errstr<<"New source list:\n";
		errstr<<sl<<endl
			  <<Sas::message(Msg::App, Msg::Silent);

		unsigned N = _slRef->size();
		double sumsq = 0.;
		for (unsigned n=0; n<N; ++n) {
			//	find all sources whose counts do not differ by more than +-7%
			const Source &refSrc = (*_slRef)[n];
			double refCounts = refSrc.counts();
			vector<unsigned> indx;
			unsigned M=sl.size();
			for (unsigned m=0; m<M; ++m)
				if (sl[m].counts() >= .93 * refCounts &&
					sl[m].counts() <= 1.07 * refCounts)
					indx.push_back(m);
			if (indx.size() == 0)
				errstr<<"Could not find reference source #"<<n<<" in source "
						"list."<<Sas::error("SourceMismatch");
			unsigned sn = indx[0];
			if (indx.size() > 1) {
				errstr<<"There are "<<indx.size()<<" sources which have the "
						"same number of counts within +-7% - choosing the one "
						"which is closest"
					  <<Sas::warning("PossibleSourceConfusion");
				double s = 1.e20;
				for (unsigned k=0; k<M; ++k) {
					double delta = refSrc.angularSeparation(sl[k]);
					if (delta < s) {
						sn = k;
						s = delta;
					}
				}
			}
			double deltaRa = cos(refSrc.dec())*(refSrc.ra() - sl[sn].ra()) *
								RAD_TO_DEG * 3600.,
				   deltaDec = (refSrc.dec() - sl[sn].dec()) * RAD_TO_DEG*3600.;
			sumsq += deltaRa*deltaRa + deltaDec*deltaDec;
		}

		errstr<<"Iteration "<<_iter-1<<" - sum of squares: "
			  <<setprecision(5)<<sumsq<<" arcsec - mean angular separation: "
			  <<setprecision(3)<<sqrt(sumsq/DBL(N))<<" arcsec\n"
			  <<Sas::message(Msg::App, Msg::Silent);
		if (_iter>=_maxIter)
			errstr<<"Maximum number of iterations ("<<_maxIter<<") reached"
				  <<Sas::error("MaxIterationsReached");

		RETURN sumsq;
	} // angularSeparation()

	~SourceCorrelator() { PROC ENDPROC }

private:
	static SourceList *_slRef;
	static unsigned	_iter;
	static unsigned _maxIter;
	static double	_phi, _theta, _psi;
	static bool		_runShowDetect;
};

SourceList *SourceCorrelator::_slRef = 0;
unsigned SourceCorrelator::_maxIter = 0;
unsigned SourceCorrelator::_iter = 0;
double SourceCorrelator::_phi = 0;
double SourceCorrelator::_theta = 0;
double SourceCorrelator::_psi = 0;
bool SourceCorrelator::_runShowDetect = true;

void epicbscalgen()
{
	PROC

	string eventTable(stringParameter("eventtable"));

	double phi0		= (realParameter("phi") + 180.*3600.)/3600. * DEG_TO_RAD,
		   theta0	= realParameter("theta")/3600. * DEG_TO_RAD,
		   psi0		= realParameter("psi") * DEG_TO_RAD;
	bool runShowDetect = booleanParameter("runshowdetect");
	double tol1 = realParameter("tol1"),
		   tol2 = realParameter("tol2");
	int	   maxIter  = intParameter("maxiter");

	SetBlock		evl(eventTable);
	Xmm::Instrument instrument;
    if (evl.table()->hasAttribute("INSTRUME"))
		instrument = Xmm::toInstrument(
						evl.table()->attribute("INSTRUME")->asString());
	else
		instrument = Xmm::toInstrument(
						evl.table()->parent()->attribute("INSTRUME")->
							asString());
	dataSetServer->close(evl.dataSet());

	SourceList *slRef = new SourceList(eventTable,
									   instrument,
									   stringParameter("xcolumn"),
									   stringParameter("ycolumn"),
									   intParameter("binx"),
									   intParameter("biny"),
									   realParameter("sourceradius"),
									   intParameter("nruns"),
									   realParameter("likemin"));

	SourceCorrelator sc(slRef, runShowDetect);
	SourceCorrelator::phi(phi0);
	SourceCorrelator::theta(theta0);
	SourceCorrelator::psi(psi0);	

	double phi = phi0,
		   theta = theta0,
		   psi = psi0;

	errstr<<"\n---------- Boresight Angle Computation: Optimization of "
			"phi/theta ----------\n"
		  <<Sas::message(Msg::App, Msg::Silent);

	int	   nVars1 = 2;
	SourceCorrelator::maxIter(maxIter);

	//	define the initial simplex
	double **p1 = Numeric::numMatrix(1, nVars1+1, 1, nVars1);

	//	first simplex vertex
	p1[1][1] = phi0;
	p1[1][2] = theta0;

	//	second simplex vertex
    p1[2][1] = phi0+50./3600.*DEG_TO_RAD;	//	phi		+50" off
	p1[2][2] = theta0+50./3600.*DEG_TO_RAD;	//	theta	+50" off

	//	third simplex vertex
	p1[3][1] = phi0-50./3600.*DEG_TO_RAD;	//	phi		-50" off
	p1[3][2] = theta0+50./3600.*DEG_TO_RAD;	//	theta	+50" off

	//	compute function for initial simplex vertices
	double *y1 = Numeric::numVector(1, nVars1+1);
	for (int n=1; n<=nVars1+1; ++n)
		y1[n] = SourceCorrelator::optimizePhiTheta(p1[n]);

	//	kick off minimization
	int nfunc1 = 0;
	Numeric::simplex(p1, y1, nVars1, tol1,
					 SourceCorrelator::optimizePhiTheta, &nfunc1);


	//
	//	stage II: optimization of phi/theta/psi
	//
	errstr<<"\n---------- Boresight Angle Computation: Optimization of "
			"phi/theta/psi ----------\n"
		  <<Sas::message(Msg::App, Msg::Silent);

	int	   nVars2 = 3;
	SourceCorrelator::maxIter(maxIter);

	//	define the initial simplex
	double **p2 = Numeric::numMatrix(1, nVars2+1, 1, nVars2);

	//	first simplex vertex
	p2[1][1] = p1[1][1];
	p2[1][2] = p1[1][2];
	p2[1][3] = psi0;

	//	second simplex vertex
    p2[2][1] = p1[2][1];
	p2[2][2] = p1[2][2];
	p2[2][3] = psi0+.5*DEG_TO_RAD;				//	psi		+1/2 def off

	//	third simplex vertex
	p2[3][1] = p1[3][1];
	p2[3][2] = p1[3][2];
	p2[3][3] = psi0-.5*DEG_TO_RAD;				//	psi		-1/2 def off

	//	fourth simplex
	p2[4][1] = p1[1][1] + 5./3600.*DEG_TO_RAD;	//	phi		+5" off
	p2[4][2] = p1[1][2] - 5./3600.*DEG_TO_RAD;	//	theta	-5" off
	p2[4][3] = psi0-.3*DEG_TO_RAD;				//	psi		-1/3 def off

	//	compute function for initial simplex vertices
	double *y2 = Numeric::numVector(1, nVars2+1);
	for (int n=1; n<=nVars2+1; ++n)
		y2[n] = SourceCorrelator::optimizePhiThetaPsi(p2[n]);

	//	kick off minimization
	int nfunc2 = 0;
	Numeric::simplex(p2, y2, nVars2, tol2,
					 SourceCorrelator::optimizePhiThetaPsi, &nfunc2);

	//	if we arrive hear we have a solution - optmized phi/theta are
	//	center of gravity of simplex
	phi = (p2[1][1]+p2[2][1]+p2[3][1]+p2[4][1])/4.,
	theta = (p2[1][2]+p2[2][2]+p2[3][2]+p2[4][2])/4.,
	psi = (p2[1][3]+p2[2][3]+p2[3][3]+p2[4][3])/4.;

	EulerAngles	angles(phi, theta, psi, EulerAngles::Euler321);
	errstr<<"Making a last iteration ...\n"
		  <<Sas::message(Msg::App, Msg::Silent);
	SourceList sl(slRef->eventListName(), slRef->instrument(),
				  slRef->xAxis(), slRef->yAxis(), angles,
				  slRef->binX(), slRef->binY(), slRef->sourceRadius());
	errstr<<sl<<endl<<Sas::message(Msg::App, Msg::Silent);

	errstr<<"A unique solution has been found. The optimized boresight "
			"misalignment angles in stage 2 are:"<<endl
		  <<"   phi   = 180 deg + "<<(phi*RAD_TO_DEG-180.)*3600.<<" arcsec\n"
		  <<"   theta = "<<theta*RAD_TO_DEG*3600.<<" arcsec\n"
		  <<"   psi   = "<<psi*RAD_TO_DEG<<" deg\n\n"
		  <<"The corresponding entry for the CCF Boresight constituent is:\n\n"
		  <<Xmm::toString(instrument)<<' '<<phi<<' '<<theta<<' '<<psi
		  <<" 0. 0. 0.\n"
		  <<Sas::message(Msg::App, Msg::Silent);

	errstr<<"You accomplished a lot today - take the rest of the day off!"
		  <<Sas::message(Msg::App, Msg::Silent);

	delete slRef;

	ENDPROC
} // epicbscalgen()
