//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	rgslincoordgrcal --- generate ASCII input file for LinCoord creation
//						 from ground-cal measurements
//
//	DESCRIPTION:
//		This little program creates the ASCII input file `lincoord.dat'
//		 needed for	the generation of RGS::LinCoord (see [3]) via the
//		 ccfmake/deceit mechanism. Invocation:
//
//			rgslincoordgrcal
//
//		Input:
//			o CCD alignment data are read from stdin in the format
//			  (see e.g. rgs1_align.dat)
//					ccdX "A|B|C|D" X Y Z
//			o lines which do not match this pattern are skipped
//			o A|B|C|D label the corner points of the CCD according to [2]
//			o X/Y/Z are the positions of the CCD in the SRON camera system [2]
//
//		Output:
//			o written to stdout
//			o format:
//				ccdid nodeid x y z dx dy dz phi theta psi dphi dtheta dpsi
//					rawX0 rawY0
//			o ccdid/nodeid: 1-9/0-1
//			o x/y/z: coordinates of the CCD pixel rawX0/rawY0 in the CAL
//			  CAMCOORD1 frame
//			o dx/dy/dz: error of x/y/z
//			o phi/theta/psi: 3-1-3 Euler angles of chips in CAMCOORD1
//			o dphi/dtheta/dpsi: error of phi/theta/psi
//			o rawX0/rawY0: CCD reference point
//
//		The employed algorithm should be obvious from the below code comments
//
//	CHANGE HISTORY:
//		UL, 1999-07-01: 1.0.0: - first working version
//		UL, 1999-09-30: 2.0.0: - X0/Y0/Z0 now correspond to center of
//								 pixel 512/192 (1x1) on each chip
//
//	AUTHOR:
//		U. Lammers, Fri Jul  2 11:54:39 MET DST 1999
//
//	REFERENCES:
//		[1]: Instrument Science Verification Review, RGS-ROL-PM-144, v1.0,
//			 p. 79
//		[2]: SRON drawings "ALIGN. CCD's FINAL ASSY FM1/2"; Drawing Nrs.
//			 322-G-2119/2120, sheet 2
//		[3]: CCF ICD
//		[4]: SAS CAL API Guide
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/ccftools/rgs/rgslincoordgrcal.cc,v 1.1 2000/04/27 13:30:01 ulammers Exp $
//

#include <iostream>
#include <vector>
#include "CcdDataReader.h"
#include "LinCoord.h"
#include "CoordinateFrame.h"

const double PIXELS_X	= 1024.;	//	pixels in dispersion direction
const double PIXELS_Y	= 384.;		//	pixels in cross-dispersion direction
									//	(only image section)
const unsigned	M_CCD	= 5;		//	M-point lies on this CCD in the
									//	geometrical	center of the image section
const double SIZE_X		= 27.65;	//	physical width [mm]
const double SIZE_Y		= 10.37;	//	physical height of image section [mm]

const double PIXEL_SIZE_X = SIZE_X/PIXELS_X;
const double PIXEL_SIZE_Y = SIZE_Y/PIXELS_Y;

//
//	CcdFrame class
//
class CcdFrame : public CcdData, public CoordinateFrame {
public:
	static const double ORTHO_EPS = 1.e-14;	//	scalar product of local
											//	chip +X and +Y must be smaller
											//	than ORTHO_EPS to be considered
											//	orthogonal
	static const unsigned MAX_ITER	= 500;	//	max number of iterations
											//	in othogonalization of X/Y

	CcdFrame(const CcdData &ccdData) : CcdData(ccdData)
	{
		unsigned i = 0;
		CartesianVector x = B - A,	//	this becomes the chip +X axis
						y = C - A;	//	this becomes the chip +Y axis
		x.normalize();
		y.normalize();
		CartesianVector z = crossProd(x, y);	//	this becomes the chip +Z
		z.normalize();							//	axis

		double xTy = fabs(x * y);
		while (xTy > ORTHO_EPS && i++<MAX_ITER) {
			//
			//	X/Y are not necessarily orthogonal, try to achieve this
			//	by below iterations
			//
			CartesianVector newX(crossProd(y, z));
			newX.normalize();
			CartesianVector tmp = .5 * x + .5 * newX;
			tmp.normalize();
			y = crossProd(z, tmp);
			x = newX;
			z = crossProd(x, y);
			z.normalize();

			xTy = fabs(x * y);
//			cerr<<"iter "<<i<<" "<<setprecision(15)<<xTy<<endl;
		}

		if (i >= MAX_ITER)
			errstr<<"Could not construct orthogonal system from corner points "
					"of chip #"<<id<<": "<<A<<"/"<<B<<"/"<<C<<"/"<<D
				  <<" after "<<MAX_ITER<<" iterations - something is wrong\n"
				  <<Sas::error("DataIntegrityError");

		CartesianVector cen(centerImageSection());	//	this is the M-point
		CoordinateFrame &fr = *this;
		fr = CoordinateFrame(cen, x, y);
	}

private:
	CartesianVector centerOfGravity() const
	{
		return .25 * (A + B + C + D);
	}

	CartesianVector centerImageSection() const
	{
		CartesianVector O = centerOfGravity(),	//	center of the chip
		  P = .5 * (D + C) - O;	                //	vector going from chip
												//	center to mid-point between
												//	D/C in the image section
		P.normalize();		//	unit vector pointing to M
		P *= .5 * SIZE_Y;	//	vector from center to M

		return O + P;
	}
};


int main(int argc, char **argv)
{
	if (--argc >0)
		errstr<<"Usage: "<<argv[0]<<endl
			  <<Sas::fatal("IllegalInvocation");

	CcdDataReader dr(cin);
	vector<CcdFrame> ccdFrame;

	//	read CCD data
	while (dr.hasMore())
		ccdFrame.push_back(*(new CcdFrame(dr.get())));

	cerr<<"Read data for "<<ccdFrame.size()<<" CCDs\n";
	if (M_CCD > ccdFrame.size() || M_CCD <= 0)
		errstr<<"Cannot choose CCD #"<<M_CCD<<" as center"
			  <<Sas::error("OutOfBounds");

	//	get origin of central CCD
	CartesianVector originCenter = ccdFrame[M_CCD-1].origin(),
					pixelCenter0(-.5*PIXEL_SIZE_X, .5*PIXEL_SIZE_Y, 0.),
					pixelCenter1(.5*PIXEL_SIZE_X, .5*PIXEL_SIZE_Y, 0.);

	//	loop over all CCDs
	LinCoord::header(argv[0]);
	for (unsigned n=0; n<ccdFrame.size(); ++n) {
		CoordinateFrame fr = ccdFrame[n];
		CartesianVector o = fr.origin() - originCenter;

		//	node 0
		//	construct attitude matrix A of chip #n to derive the 3-1-3 Euler
		//	angles; the rows of A are the unit vectors of the body frame
		AttitudeMatrix a0(fr.plusX(), fr.plusY(), fr.plusZ());
		EulerAngles ea0 = a0.toEulerAngles();
		LinCoord::lineout(n+1, 0, o+pixelCenter0, ea0, PIXELS_X/2, PIXELS_Y/2);

		//	node 1
		AttitudeMatrix a1(-fr.plusX(), fr.plusY(), -fr.plusZ());
		EulerAngles ea1 = a1.toEulerAngles();
		LinCoord::lineout(n+1, 1, o+pixelCenter1, ea1, PIXELS_X/2, PIXELS_Y/2);
	}

	return 0;
}
