//
//	rgslincoord: compute the nominal 3-1-3 Euler angles + position of
//				 reference pixels of RFC chips in the S/C frame
//
//	UL, Thu Mar 11 15:15:13 MET 1999
//
//	REFERENCES:
//		[1]: RS-PX-0019, Figure 2.1
//		[2]: Instrument Science Verification Review, RGS-ROL-PM-144
//		[2]: Informal TN, Author: UL
//
#include <cmath>
#include <iomanip>
#include <iostream>

//
//	math constants
//
const double PI			= acos(-1.),
			 PI_HALF	= .5 * PI,
			 DEG_TO_RAD	= PI/180.;

template<class T>
T SQR(T arg) { return arg*arg; }

//
//	RGA/RFC constants
//
const double
	GX		= 0.,		//	x/y/z components of location vector of
	GY		= 0.,		//	center of RGA in S/C frame [mm]
	GZ		= 0.,
	R		= 3352.64,	//	radius of Rowland circle [mm]
	GAMMA	= 2.275068,	//	angle between optical axis and O-G with
						//	O being the center of the Rowland circle [deg]
	ETA		= 7.333,	//	angle between optical axis and M-O with M being
						//	the geometrical center of the centr. RFC chip [deg]
	W		= 27.65,	//	width of active chip area [mm]
	H		= 10.37,	//	height of active image area of RFC chips [mm]
	G		= .54975;	//	gap between adjacent chips [mm]

const int CCDS			= 9,	//	number of chips in RFC
		  CENTER_CCD	= 5,	//	number of central one
		  NODES			= 2,	//	number of nodes
		  PIXELSX		= 1024;	//	number of pixels in dispersion dir.

int main()
{
	const double coseta		= cos(ETA * DEG_TO_RAD),
				 sineta		= sin(ETA * DEG_TO_RAD),
				 rcosgamma	= R * cos(GAMMA * DEG_TO_RAD),
				 rsingamma	= R * sin(GAMMA * DEG_TO_RAD);

	cout<<"CCD_ID NODE_ID X0 Y0 Z0 EULER_PHI EULER_THETA EULER_PSI "
		  "RAWX0 RAWY0\n";
	cout.precision(15);
	for (int node=0; node<NODES; ++node)
		for (int ccd=1; ccd<=CCDS; ++ccd) {
			double x = (ccd-CENTER_CCD)*(W+G)+.5*W,
				   y = R - sqrt(SQR(R) - SQR(ccd-CENTER_CCD)*SQR(W+G)),

				   pix = GX + rcosgamma - (sineta * x + coseta * y),
				   piy = GY + .5 * H,
				   piz = GZ - rsingamma - (coseta * x - sineta * y),

				   phi = PI_HALF,
				   theta = node == 0 ? PI_HALF + DEG_TO_RAD * ETA
									 : -(PI_HALF - DEG_TO_RAD * ETA),
				   psi = PI_HALF;

			unsigned refx = node == 0 ? 1 : PIXELSX,
					 refy = 1;

			cout<<ccd<<" "<<node<<" "
				<<setprecision(8)<<pix<<" "<<piy<<" "<<piz<<" "
				<<setprecision(15)
				<<phi<<" "<<theta<<" "<<psi<<" "<<refx<<" "<<refy<<endl;
		}

	return 0;
}
