//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	siamgen.cc	--- generate the SIAM matrix accorind to [1]
//
//	DESCRIPTION:
//		This task generates the SIAM matrices for the instruments
//		EMOS1, EMOS2, EPN, RGS1, RGS2, and OM from the contents of the
//		CCF Boresight constituent. The task has no mandatory parameters
//		and writes the output matrices to <stdout> in the format specified
//		in [1]. In order to provide the SIAM matrices to FDS according
//		to [1] redirect <stdout> into a file and wrap it up with a
//		corresponding XFTS wrapper.
//
//	OPTIONS:
//		--comment=<string>				: comment string for the first SIAM
//										  record
//		--withcaltime<inst>=true|false	: boolean switch determining whether
//										  or not the time of boresight
//										  calibration for instrument <inst>=
//										  emos1|emos2|epn|rgs1|rgs2|om
//										  is explicitely given in the
//										  caltime<inst> argument; if false,
//										  this time which will appear in the
//										  first line of the respective SIAM
//										  entry is taken as the current time
//		--caltime<inst>=<yyyy-mm-ddThh:mm:ss>
//										  time of the borsight calibration
//										  of instrument <inst>
//
//	CHANGE HISTORY:
//		UL, 1999-09-23: 1.0.0: - first version
//		UL, 1999-11-11: 1.1.0: - brought in line with cal-3.32
//		UL, 1999-10-30: 1.2.0: - brought in line with SIAM ICD issue 1.2:
//								 the constraints that only two of the
//								 angles are independent from each other
//								 has been removed; now there can also
//								 be a proper calibration around the
//								 pointing axis
//
//	AUTHOR:
//		U. Lammers, Thu Sep 23 13:53:49 MET DST 1999
//
//	REFERENCES:
//		[1]: XMM Star Tracker/Instrument Alignment Calibration - Interface
//			 Control Document, Issue 1.2, R. Mugellesi, ESOC/ESA
//			 Online: ftp://astro.estec.esa.nl/pub/XMM/documents/
//					 siam_icd.pdf
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/ccftools/siamgen/src/siamgen.cc,v 1.3 1999/11/30 13:39:28 ulammers Exp $
//
#include <iostream>
#include <iomanip>
#include <strstream>
#include <cstdio>	//	sprintf: formatted I/O with C++ streams is just
					//	too painful!
#include <cmath>	//	fabs
#include <map>

#include "common.h"
#include "param.h"
#include "Tracer.h"
#include "stringutil.h"
#include "STime.h"
#include "Xmm.h"
#include "Celestial.h"

#include "CalServer.h"
#include "Boresight.h"

#include "siamgen.h"

namespace {

enum Flush { FLUSH } flushit;

class SiamRecord : public strstream {
public:
	SiamRecord(unsigned length = 64, ostream &os = cout)
		: _len(length), _os(os) { PROC ENDPROC }

	friend SiamRecord &operator<<(SiamRecord &, Flush);

	~SiamRecord() { PROC ENDPROC }

private:
	unsigned	_len;
	ostream		&_os;
};

template<class T>
SiamRecord &operator<<(SiamRecord &r, T t)
{
	PROC
	strstream &s = r;
	s<<t;
	RETURN r;
}

SiamRecord &operator<<(SiamRecord &r, Flush)
{
	PROC

	r<<ends;
	string s = r.str() + string(r._len-1, ' ');
	r._os<<s.substr(0, r._len-1)<<endl;

	return r;
} // operator<<


class InstIds : public map<Xmm::Instrument, string, less<Xmm::Instrument> > {
public:
	InstIds()
	{	PROC

		map<Xmm::Instrument, string, less<Xmm::Instrument> > &m = *this;
		m[Xmm::EMOS1]	= "E1";
		m[Xmm::EMOS2]	= "E2";
		m[Xmm::EPN]		= "E3";
		m[Xmm::RGS1]	= "R1";
		m[Xmm::RGS2]	= "R2";
		m[Xmm::OM]		= "O1";

		ENDPROC
	}
	~InstIds() { PROC ENDPROC }
} instIds;

template<class T>
bool isAlmostNull(T v) { return fabs(v) < 1.e-6; }

class SIAMmatrix : public AttitudeMatrix {
public:
	SIAMmatrix(const AttitudeMatrix &a) : AttitudeMatrix(a) {}

	bool isGood() const {
		//
		//	S/C +Z axis must lie in +X/+Z plane of boresight frame
		//	see Figure 2 in [1]
		//
		CartesianVector zInInsBor(*this * CartesianVector(0., 0., 1.));

		//	S/C +Z vector in instrument boresight frame (IBF) and IBF
		//	+Y axis must be orthogonal
		//	return isAlmostNull(zInInsBor * CartesianVector(0., 1., 0.));
		//	PLEASE NOTE: This was changed from [1] issue 1.0 to 1.2: All
		//				 three BS anglea are now independent from each other,
		//				 i.e., the S/C +Z axis has no longer to lie in the
		//				 instrument +X/+Z plane
		return true;
	}
};

} // unnamed namespace


void siamgen()
{
	ios::sync_with_stdio();		//	we mix C++ streams + C-stdio

	//	write header record
	SiamRecord headerRecord(65);
	STime ct(STime::currentTime());
	headerRecord<<ct.toString(0)<<"Z RPRT  "<<instIds.size()<<' '
				<<stringParameter("comment")<<flushit;

	//	loop over all instruments
	for (unsigned n=0; n<xmmPayLoad.instruments().size(); ++n) {
		XmmInstrument *ins = xmmPayLoad.instruments()[n];
        if (ins->id() == Xmm::ERM)
            continue;	//	no SIAM entry for ERM

		//	set cal state and retrieve boresight matrix
		calServer.setInstrument(ins->id());
		Boresight *bs = calServer.getAtom(bs);

		//	write first line of SIAM block #n
		SiamRecord line1(24);
		string caltime = "caltime" + lowercase(Xmm::toString(ins->id()));
		STime t(ct);
		if (booleanParameter("with" + caltime)) {
			STime tnew(stringParameter(caltime));
			t = tnew;
		}
		line1<<instIds[ins->id()]<<" "<<t.toString(0)<<'Z'<<flushit;

		SIAMmatrix sm(bs->matrix(0.));
		//	do not need to check orthogonality because that's guaranteed
		//	by AttitudeMatrix construction through Euler angles;
		//	but we need to check if S/C +Z axis is in +X/+Z of IBF
		//
		if (sm.isGood())
			for (unsigned i=0; i<3; ++i) {
				SiamRecord rec(63);
				char s[21];
				for (unsigned j=0; j<3; ++j) {
					sprintf(s, "%-20.17f", (*sm.data())[i][j]);
					rec<<s<<' ';
				}
				rec<<flushit;
			}
		else
			errstr<<"SIAM matrix for instrument "<<instIds[ins->id()]
				  <<" is invalid - +Z vector of S/C frame is not in "
				  <<"+X/+Z plane of instrument boresight frame (see "
					"figure 2 in SIAM ICD"
				  <<Sas::error("WrongSIAMmatrix");
	}
}
